#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

from __future__ import print_function

import os
import shutil
import subprocess
import subprocess as sp
import tarfile
import tempfile
import zipfile
from time import sleep


class ZipFile(zipfile.ZipFile):
    """ Custom ZipFile class handling file permissions. """
    def _extract_member(self, member, targetpath, pwd):
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        targetpath = super()._extract_member(member, targetpath, pwd)

        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(targetpath, attr)
        return targetpath


class TypeDBServer(object):

    def __init__(self, distribution_location):
        self.__distribution_location = distribution_location
        self.__distribution_root_dir = None
        self.__unpacked_dir = None
        self.__pid = None

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.stop()

    def start(self):
        if not self.__unpacked_dir:
            self._unpack()
        popen = subprocess.Popen(["typedb", "server"],
                                 cwd=os.path.join(self.__unpacked_dir, self.__distribution_root_dir))
        sleep(5)
        self.__pid = popen.pid

    def stop(self):
        sp.check_call([
            "kill", f"{self.__pid}"
        ], cwd=os.path.join(self.__unpacked_dir, self.__distribution_root_dir))
        shutil.rmtree(self.__unpacked_dir)
        self.__pid = None

    def _unpack(self):
        self.__unpacked_dir = tempfile.mkdtemp(prefix='typedb')
        with tarfile.open(self.__distribution_location) as tf:
            tf.extractall(self.__unpacked_dir)
            self.__distribution_root_dir = os.path.commonpath(tf.getnames()[1:])

    @property
    def typedb_binary_location(self):
        return os.path.join(self.__unpacked_dir, self.__distribution_root_dir)



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

from typing import Sequence, Callable, Optional

import networkx as nx
from torch_geometric.utils import from_networkx
from typedb.client import TypeDBSession, TypeDBOptions, TransactionType

from typedb_ml.networkx.queries_to_networkx import build_graph_from_queries


class DataSet:
    """
    A DataSet to lazily load graphs based on queries from TypeDB and some arbitrary id.
    """

    def __init__(
            self,
            indices: Sequence,
            node_types,
            edge_type_triplets,
            queries_for_id: Callable,
            session: Optional[TypeDBSession] = None,
            infer: bool = True,
            transform: Optional[Callable[[nx.Graph], nx.Graph]] = None,
    ):
        self._indices = indices
        self._node_types = node_types
        self._edge_type_triplets = edge_type_triplets
        self.queries_for_id = queries_for_id
        self._infer = infer
        self._transform = transform
        self.session = session

    def __len__(self):
        return len(self._indices)

    def __getitem__(self, idx):
        id = self._indices[idx]
        print(f"Fetching graph for id: {id}")
        queries = self.queries_for_id(id)

        options = TypeDBOptions.core()
        options.infer = self._infer

        with self.session.transaction(TransactionType.READ, options=options) as tx:
            # Build a graph from the queries
            graph = build_graph_from_queries(queries, tx)
        graph.name = id
        if self._transform:
            graph = self._transform(graph)
        data = from_networkx(graph)
        data.concepts_by_type = graph.concepts_by_type
        return data, self.node_type_indices(graph), self.edge_type_indices(graph)

    def node_type_indices(self, graph):
        indices = []
        for _, data in graph.nodes(data=True):
            indices.append(self._node_types.index(data['type']))
        return indices

    def edge_type_indices(self, graph):
        indices = []
        for src, dst, data in graph.edges(data=True):
            indices.append(self._edge_type_triplets.index((graph.nodes[src]["type"], data["type"], graph.nodes[dst]["type"])))
        return indices



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#
from typedb.api.concept.type.attribute_type import AttributeType


class MockConcept:
    def __init__(self, iid):
        self.iid = iid


class MockType(MockConcept):
    def __init__(self, iid, label, base_type):
        super().__init__(iid)
        self._label = Label(label)
        assert base_type in {'ENTITY', 'RELATION', 'ATTRIBUTE'}
        self.base_type = base_type

    def get_label(self):
        return self._label


class Label:
    def __init__(self, name):
        self._name = name

    def name(self):
        return self._name


class ValueType:
    def __init__(self, name):
        self.name = name


class MockAttributeType(MockType):
    def __init__(self, iid, label, base_type, value_type):
        super().__init__(iid, label, base_type)
        assert value_type in {AttributeType.ValueType.OBJECT,
                              AttributeType.ValueType.STRING,
                              AttributeType.ValueType.LONG,
                              AttributeType.ValueType.DOUBLE,
                              AttributeType.ValueType.DATETIME,
                              AttributeType.ValueType.BOOLEAN}
        self._value_type = value_type

    def get_value_type(self):
        return self._value_type


class MockThing(MockConcept):
    def __init__(self, iid, type):
        super().__init__(iid)
        assert(isinstance(type, MockType))
        self._type = type

    def get_type(self):
        return self._type

    def get_iid(self):
        return self.iid

    def is_entity(self):
        return 'ENTITY' in self._type.base_type

    def is_attribute(self):
        return 'ATTRIBUTE' in self._type.base_type

    def is_relation(self):
        return 'RELATION' in self._type.base_type

    @property
    def base_type(self):
        return self._type.base_type


class MockAttribute(MockThing):
    def __init__(self, iid, value, type):
        super().__init__(iid, type)
        self._value = value
        assert isinstance(type, MockAttributeType)

    def get_value(self):
        return self._value



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import networkx as nx


class QueryGraph(nx.MultiDiGraph):
    """
    A custom graph to represent a query. Has additional helper methods specific to adding TypeQL patterns.
    """

    def add_vars(self, vars):
        """
        Add variables, stored as nodes in the graph
        Args:
            vars: String variables

        Returns:
            self
        """
        for var in vars:
            self.add_node(var)
        return self

    def add_has_edge(self, owner_var, attribute_var):
        """
        Add a "has" edge to represent ownership of an attribute
        Args:
            owner_var: The variable of the owner
            attribute_var: The variable of the owned attribute

        Returns:
            self
        """
        self.add_edge(owner_var, attribute_var, type='has')
        return self

    def add_role_edge(self, relation_var, roleplayer_var, role_label):
        """
        Add an edge to represent the role a variable plays in a relation
        Args:
            relation_var: The variable of the relation
            roleplayer_var: The variable of the roleplayer in the relation
            role_label: The role the roleplayer plays in the relation

        Returns:
            self
        """
        self.add_edge(relation_var, roleplayer_var, type=role_label)
        return self


class Query:
    def __init__(self, graph: nx.MultiDiGraph, string: str):
        self.graph = graph
        self.string = string

    def __str__(self):
        return self.string



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

from typedb.api.concept.type.attribute_type import AttributeType


class Thing:

    VALUE_TYPES = (
        AttributeType.ValueType.OBJECT,
        AttributeType.ValueType.STRING,
        AttributeType.ValueType.LONG,
        AttributeType.ValueType.DOUBLE,
        AttributeType.ValueType.DATETIME,
        AttributeType.ValueType.BOOLEAN
    )

    def __init__(self, iid, type_label, base_type, value_type=None, value=None):
        self._hash = None
        self.iid = iid
        self.type_label = type_label
        self.base_type = base_type

        # If the thing is an attribute
        self.value_type = value_type
        self.value = value

        # TODO Make attribute a separate class
        if self.base_type == 'attribute':
            if self.value_type is None:
                raise ValueError('Attribute value_type must be provided')
            if self.value is None:
                raise ValueError('Attribute value must be provided')

    def __str__(self):
        string = f'<{self.type_label}, {self.iid}'
        if self.base_type == 'attribute':
            string += f': {self.value}'
        return string + '>'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, self.__class__):
            if self.iid == other.iid:
                assert self.__dict__ == other.__dict__
                return True
            else:
                return False
        return NotImplemented

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        x = self.__eq__(other)
        if x is not NotImplemented:
            return not x
        return NotImplemented

    def __hash__(self):
        """Overrides the default implementation"""
        if self._hash is None:
            self._hash = hash(self.iid)
        return self._hash


def build_thing(typedb_thing):

    iid = typedb_thing.get_iid()
    type_label = typedb_thing.get_type().get_label().name()
    if typedb_thing.is_entity():
        base_type = "entity"
    elif typedb_thing.is_relation():
        base_type = "relation"
    elif typedb_thing.is_attribute():
        base_type = "attribute"
    else:
        raise RuntimeError("Unexpected Concept")

    if base_type == 'attribute':
        value_type = typedb_thing.get_type().get_value_type()
        assert value_type in Thing.VALUE_TYPES
        value = typedb_thing.get_value()

        return Thing(iid, type_label, base_type, value_type, value)

    return Thing(iid, type_label, base_type)



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import subprocess as sp
from enum import Enum


class FileType(Enum):
    Schema = "schema"
    Data = "data"


def load_typeql_file(typedb_binary_location: str, database_name: str, typeql_file_path: str,
                     file_type: FileType):
    """
    Load a file into a TypeDB database

    Args:
        typedb_binary_location: the location of TypeDB
        database_name: The name of the TypeDB database to load into
        typeql_file_path: The path to the file to load
        file_type: The content of the file and therefore the transaction type to use, either schema or data
    Returns:
        None
    """
    sp.check_call([
        './typedb',
        'console',
        f'--command=transaction {database_name} {file_type.value} write',
        f'--command=source {typeql_file_path}',
        f'--command=commit'
    ], cwd=typedb_binary_location)



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#
from typedb.api.connection.transaction import TransactionType


def get_thing_types(session):
    """
    Get all schema types, excluding those for implicit attribute relations and base types
    Args:
        session: TypeDB session

    Returns:
        TypeDB types
    """
    with session.transaction(TransactionType.READ) as tx:
        schema_concepts = tx.query().match("match $x sub thing;")
        thing_types = [schema_concept.get('x').get_label().name() for schema_concept in schema_concepts]
    [thing_types.remove(el) for el in ['thing', 'relation', 'entity', 'attribute']]
    return thing_types


def get_role_triplets(tx):
    """
    Get triples of all schema roles and the relation and roleplayer they connect
    Args:
        tx: TypeDB transaction

    Returns:
        TypeDB role triples
    """
    role_triples = []
    schema_concepts = tx.query().match("match $rel sub relation; $rel relates $r; $rp plays $r;")
    for answer in schema_concepts:
        relation = answer.get('rel').get_label().name()
        roles = [r.get_label().name() for r in answer.get('r').as_remote(tx).get_supertypes()]
        roles.remove('role')
        player = answer.get("rp").get_label().name()
        for role in roles:
            role_triples.append((relation, role, player))
    return role_triples


def get_has_triplets(tx):
    """
    Get triples of all ownerships: the owner type and owned attribute
    Args:
        tx: TypeDB transaction

    Returns:
        TypeDB ownership triples
    """
    has_triples = []
    schema_concepts = tx.query().match("match $owner sub thing, owns $owned;")
    for answer in schema_concepts:
        owner = answer.get('owner').get_label().name()
        owned = answer.get('owned').get_label().name()
        has_triples.append((owner, "has", owned))
    return has_triples


def get_edge_type_triplets(session):
    with session.transaction(TransactionType.READ) as tx:
        edge_types = get_role_triplets(tx) + get_has_triplets(tx)
    return edge_types


def reverse_edge_type_triplets(edge_types):
    reversed_edge_triples = []
    for edge_from, edge, edge_to in edge_types:
        reversed_edge_triples.append((edge_to, f"rev_{edge}", edge_from))
    return reversed_edge_triples



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#
import networkx as nx

from typedb_ml.networkx.iterate import multidigraph_edge_data_iterator
from typedb_ml.typedb.type import get_edge_type_triplets, reverse_edge_type_triplets


class LinkPredictionLabeller:

    def __init__(self, edge_type_to_predict):
        self.edge_type_to_predict = edge_type_to_predict

    def __call__(self, graph):
        self.label_edges(graph)
        return graph

    def label_edges(self, graph):
        for data in multidigraph_edge_data_iterator(graph):
            if data["type"] == self.edge_type_to_predict:
                data["y_edge"] = 1
            else:
                data["y_edge"] = 0


def binary_relations_to_edges(graph: nx.MultiDiGraph, binary_relation_type):
    """
    A function to convert a TypeDB relation (a hyperedge) to a directed binary edge.

    The replaced relations may not be:
    - a roleplayer in any other relations
    - own any attributes
    - have more than two roles
    and must:
    - have exactly two outgoing role edges (which can be of the same role)

    This is useful because edges are often stored as an adjacency matrix. In PyTorch Geometric, for example,
    this adjacency matrix expects binary edges. To be compatible we convert to binary edges so that when creating
    negative samples we can simply change the values of the adjacency matrix. In the case of a hyperedge we cannot
    simply do this to add negative edges.

    Args:
        graph: The graph to modify in-place
        binary_relation_type: A triple of the `(role1, relation, role2)` types to convert to a single edge labelled with `relation`

    Returns:
        Dict of edges to the node each replaces
    """
    replacement_edges = {}
    for node, node_data in graph.nodes(data=True):
        if node_data["type"] == binary_relation_type[1]:
            if len(list(graph.in_edges(node))) > 0:
                raise ValueError(
                    "The given binary relation can't be transformed into an edge because it plays a role in another "
                    "relation."
                )
            roles = list(graph.edges(node, data=True))  # equivalent to out_edges()
            assert len(roles) == 2
            new_edge_start = None
            new_edge_end = None
            for from_roleplayer, to_roleplayer, data in roles:
                if data["type"] == binary_relation_type[0]:
                    if new_edge_start is None:
                        new_edge_start = to_roleplayer
                    else:
                        # In this case the given relation type is symmetric. Therefore direction is arbitrary (but
                        # expect to add a reverse edge elsewhere).
                        assert binary_relation_type[0] == binary_relation_type[2]
                        new_edge_end = to_roleplayer
                elif data["type"] == binary_relation_type[2]:
                    new_edge_end = to_roleplayer
                else:
                    raise ValueError(
                        f"Unexpected role in relation {binary_relation_type[1]}. Expected \""
                        f"{binary_relation_type[0]}\" or \"{binary_relation_type[2]}\" but got \"{data['type']}\"."
                    )
                graph.remove_edge(from_roleplayer, to_roleplayer)
            graph.add_edge(new_edge_start, new_edge_end, type=binary_relation_type[1])
            replacement_edges[(new_edge_start, new_edge_end)] = node
    for node in replacement_edges.values():
        graph.remove_node(node)
    return graph


def binary_link_prediction_edge_triplets(session, relation_type_to_predict, types_to_ignore):
    edge_type_triplets = get_edge_type_triplets(session)
    edge_type_triplets = [e for e in edge_type_triplets if not (set(e).intersection(types_to_ignore))]
    replace_relation_with_binary_edge(edge_type_triplets, relation_type_to_predict)
    edge_type_triplets_reversed = reverse_edge_type_triplets(edge_type_triplets)
    return edge_type_triplets, edge_type_triplets_reversed


def replace_relation_with_binary_edge(edge_type_triplets, relation_type_to_predict):
    # Remove the two roles of the relation_to_predict and replace them with a binary edge that uses the name of the
    # binary relation it represents
    edge_type_triplets.remove(relation_type_to_predict[2:])
    if relation_type_to_predict[1] != relation_type_to_predict[3]:
        edge_type_triplets.remove(tuple(reversed(relation_type_to_predict[:3])))
    edge_type_triplets.append((relation_type_to_predict[0], relation_type_to_predict[2], relation_type_to_predict[4]))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from setuptools import setup
from setuptools import find_packages

packages = find_packages()

setup(
    name = "typedb-ml",
    version = "0.3.0",
    description = "A Machine Learning Library for TypeDB.",
    long_description = open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers = ["Programming Language :: Python :: 3", "Programming Language :: Python :: 3.7", "License :: OSI Approved :: Apache Software License", "Operating System :: OS Independent", "Intended Audience :: Developers", "Intended Audience :: Science/Research", "Topic :: Scientific/Engineering", "Topic :: Scientific/Engineering :: Information Analysis", "Topic :: Scientific/Engineering :: Artificial Intelligence", "Topic :: Software Development :: Libraries", "Topic :: Software Development :: Libraries :: Python Modules"],
    keywords = "machine learning logical reasoning knowledege graph typedb database graph knowledgebase knowledge-engineering",
    url = "https://github.com/vaticle/typedb-ml",
    author = "Vaticle",
    author_email = "community@vaticle.com",
    license = "Apache-2.0",
    packages=packages,
    install_requires=['absl-py==1.2.0', 'cachetools==5.2.0', 'certifi==2022.6.15', 'charset-normalizer==2.1.0', 'decorator==5.1.1', 'google-auth==2.9.1', 'google-auth-oauthlib==0.4.6', 'grpcio==1.43.0', 'idna==3.3', 'importlib-metadata==4.12.0', 'Jinja2==3.1.2', 'joblib==1.1.0', 'Markdown==3.4.1', 'MarkupSafe==2.1.1', 'networkx==2.5', 'numpy==1.21.6', 'oauthlib==3.2.0', 'pandas==1.3.5', 'protobuf==3.15.5', 'pyasn1==0.4.8', 'pyasn1-modules==0.2.8', 'pyparsing==3.0.9', 'python-dateutil==2.8.2', 'pytz==2022.1', 'requests==2.28.1', 'requests-oauthlib==1.3.1', 'rsa==4.9', 'scikit-learn==1.0.2', 'scipy==1.7.3', 'six==1.16.0', 'tensorboard==2.9.1', 'tensorboard-data-server==0.6.1', 'tensorboard-plugin-wit==1.8.1', 'threadpoolctl==3.1.0', 'torch==1.11.0', 'torch-geometric==2.0.4', 'torch-scatter==2.0.9', 'torch-sparse==0.6.14', 'tqdm==4.64.0', 'typedb-client==2.9.0', 'typedb-protocol==2.9.0', 'typing-extensions==4.3.0', 'urllib3==1.26.10', 'Werkzeug==2.1.2', 'zipp==3.8.1'],
    zip_safe=False,
)



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#
from typedb_ml.networkx.iterate import multidigraph_node_data_iterator, multidigraph_edge_data_iterator


def clear_unneeded_fields(graph):
    for node_data in multidigraph_node_data_iterator(graph):
        x = node_data["x"]
        t = node_data["type"]
        node_data.clear()
        node_data["x"] = x
        node_data["type"] = t

    for edge_data in multidigraph_edge_data_iterator(graph):
        x = edge_data["edge_attr"]
        y = edge_data["y_edge"]
        t = edge_data["type"]
        edge_data.clear()
        edge_data["edge_attr"] = x
        edge_data["y_edge"] = y
        edge_data["type"] = t
    return graph


def store_concepts_by_type(graph):
    """
    Organises concepts by type the same way `data.to_heterogeneous()` organises nodes by type. This is necessary to
    be able to map back from a `HeteroData` object to the TypeDB concepts that the nodes refer to.
    Args:
        graph: NetworkX graph to operate on
    Returns:
        The same graph, with a field `concepts_by_type` holding concepts organised by type
    """
    concepts_by_type = {}
    for node_data in multidigraph_node_data_iterator(graph):
        typ = node_data['type']
        if typ in concepts_by_type:
            concepts_by_type[typ].append(node_data['concept'])
        else:
            concepts_by_type[typ] = [node_data['concept']]
    graph.concepts_by_type = concepts_by_type
    return graph



#
#  Copyright (C) 2022 Vaticle
#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#
from typing import List

import torch
from torch.nn import Embedding

from typedb_ml.networkx.iterate import (
    multidigraph_node_data_iterator,
    multidigraph_edge_data_iterator,
)


class FeatureEncoder:
    """
    Feature encoder for NetworkX representations of TypeDB data. Type data is assumed always present for each node
    and edge, consistent with the TypeDB knowledge model. Therefore, this encoder provides a de-factor method to
    embed that type information. Supply attribute-specific encoders to handle encoding values of attributes,
    since the meaning of these values is domain-dependent.
    """

    def __init__(self, node_types, edge_types, type_encoding_size, attribute_encoders, attribute_encoding_size):
        self.node_types = node_types
        self.edge_types = edge_types
        self.attribute_encoders = attribute_encoders
        self.attribute_encoding_size = attribute_encoding_size
        self.node_type_embedding = Embedding(len(self.node_types), type_encoding_size)
        self.edge_type_embedding = Embedding(len(self.edge_types), type_encoding_size)

    def __call__(self, graph):
        self.encode_node_features(graph)
        self.encode_edge_features(graph)
        return graph

    def encode_node_features(self, graph):
        for node_data in multidigraph_node_data_iterator(graph):
            typ = node_data['type']
            if typ in self.attribute_encoders.keys():
                # Add the integer value of the category for each categorical attribute instance
                encoded_value = self.attribute_encoders[typ](node_data['value'])
            else:
                encoded_value = [0] * self.attribute_encoding_size

            type_embedding = self.node_type_embedding(torch.as_tensor(self.node_types.index(node_data['type'])))
            node_data['x'] = torch.hstack([type_embedding, torch.as_tensor(encoded_value)])\
                .cpu().detach().numpy()  # Conversion to numpy array, otherwise the graph representation breaks

    def encode_edge_features(self, graph):
        for edge_data in multidigraph_edge_data_iterator(graph):
            type_embedding = self.node_type_embedding(torch.as_tensor(self.edge_types.index(edge_data['type'])))
            edge_data['edge_attr'] = torch.hstack([type_embedding, torch.as_tensor([0] * self.attribute_encoding_size)])\
                .cpu().detach().numpy()  # Conversion to numpy array, otherwise the graph representation breaks


class CategoricalEncoder:

    def __init__(self, categories: List, attribute_encoding_size):
        self.categories = categories
        self.embedding = Embedding(len(self.categories), attribute_encoding_size)

    def __call__(self, value):
        return self.embedding(torch.as_tensor(self.categories.index(value)))


class ContinuousEncoder:
    def __init__(self, min_val, max_val, attribute_encoding_size):
        self.attribute_encoding_size = attribute_encoding_size
        self.max_val = max_val
        self.min_val = min_val

    def __call__(self, value):
        return [(value - self.min_val) / (self.max_val - self.min_val)] * self.attribute_encoding_size
